/*
 * eyeplay.c based on getjpeg
 *
 * getjpeg.c, sample tool for grabbing images from video devices
 *
 * Copyright (c) 2003 Joerg Heckenbach <joerg@heckenbach-aw.de>
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <linux/types.h>
#include <linux/videodev.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <libgen.h>
#include <asm/page.h>	

#include <linux/fb.h>

#define GETJPEG_MMAP
#define DEF_WIDTH 640
#define DEF_HEIGHT 480


#define XMD_H
#include <jpeglib.h>
#include <setjmp.h>

struct my_error_mgr {
  struct jpeg_error_mgr pub;	/* "public" fields */
  jmp_buf setjmp_buffer;	/* for return to caller */
};

typedef struct my_error_mgr * my_error_ptr;

METHODDEF(void)
my_error_exit (j_common_ptr cinfo)
{
  my_error_ptr myerr = (my_error_ptr) cinfo->err;
  (*cinfo->err->output_message) (cinfo);
  longjmp(myerr->setjmp_buffer, 1);
}

static struct jpeg_decompress_struct cinfo;
static struct my_error_mgr jerr;

static JSAMPARRAY buffer;		/* Output row buffer */
static int row_stride;		/* physical row width in output buffer */
static FILE *infile;




void
usage (char *pname)
{
  fprintf (stderr,
	   "Usage: %s <options>\n"
	   " -d <device>          video device (default: /dev/video0)\n"
	   " -s NxN               define size of the output image (default:"
	   " %dx%d)\n"
	   "Example: %s -d /dev/video0 -s 320x240 \n",
	   (char*)basename(pname), DEF_WIDTH, DEF_HEIGHT, (char*)basename(pname));
  exit (1);
}

int main(int argc, char **argv)
{
  int console_fd;
  const char *fbdev = "/dev/fb0";
  struct fb_var_screeninfo vinfo;
  struct fb_fix_screeninfo finfo;
  char *filename = "/tmp/tmp.jpg";
  short *dst;
  short *fb;
  int mapped_memlen;
  int mapped_offset;
  void *mapped_mem;
  int fb_offset;
  int i, current_line, last_line;
  unsigned char *src;
  
  int r,g,b;
  
  
  FILE *ofid;
  int ifid;
  ssize_t count;
#ifndef GETJPEG_MMAP
  unsigned char buffer[524288];
#endif
  char default_dev[] = "/dev/video0";
  char *device = default_dev;
  
  char *vmbuffer = NULL;
  unsigned short *vmpointer = NULL;
  struct video_mbuf vmbuf;
  struct video_mmap vmmap;
  int c;
  int width = DEF_WIDTH, height = DEF_HEIGHT;
  
  while ((c = getopt (argc, argv, "d:s:")) != EOF) {
    switch (c) {
    case 'd': /* change default device */
      device = optarg;
      break;
    case 's':
      sscanf (optarg, "%dx%d", &width, &height);
      break;
    default:
      usage (argv[0]);
      break;
    }
  }
  
  //	if ((ifid = open(device, O_RDWR)) < 0) {
  //		printf("Could not open device %s\n", device);
  //		return -1;
  //	}
  //	sleep(1);
  //	count = read(ifid, buffer, sizeof(buffer));
  //	close(ifid);
  //sleep(1);
  
  console_fd = open(fbdev, O_RDWR, 0);
  if ( console_fd < 0 ) {
    printf("Could not open device %s\n", fbdev);
    return(-1);
  }
  if ( ioctl(console_fd, FBIOGET_VSCREENINFO, &vinfo) < 0 ) {
    printf("Couldn't get console pixel format");
    return(-1);
  }
  if ( ioctl(console_fd, FBIOGET_FSCREENINFO, &finfo) < 0 ) {
    printf("Couldn't get console hardware info");
    return(-1);
  }
  
  mapped_offset = (((long)finfo.smem_start) -
		   (((long)finfo.smem_start)&~(PAGE_SIZE-1)));
  mapped_memlen = finfo.smem_len+mapped_offset;
  mapped_mem = mmap(NULL, mapped_memlen,PROT_READ|PROT_WRITE, MAP_SHARED, console_fd, 0);
  fb_offset = (vinfo.xres-height)+(vinfo.yres-width)*vinfo.xres;
  fb = (void *)((int )mapped_mem+fb_offset);
  
  if ((ifid = open(device, O_RDWR)) < 0) {
    printf("Could not open device %s\n", device);
    return -1;
  }
  
  
  /* setup everything */
  ioctl(ifid, VIDIOCGMBUF, &vmbuf);
  //printf("VMBUF: size=%d, frames=%d, offset[0]=%d, offset[1]=%d\n",
  //		vmbuf.size, vmbuf.frames, vmbuf.offsets[0], vmbuf.offsets[1]);
  
  vmbuffer = mmap(0, vmbuf.size, PROT_READ, MAP_SHARED, ifid, 0);
  if (!(vmbuffer))
    perror("mmap failed");
  
  vmmap.format = VIDEO_PALETTE_RGB24;
  vmmap.frame = 0;
  vmmap.width = width;
  vmmap.height = height;
  
  if (ioctl (ifid, VIDIOCMCAPTURE, &vmmap) == -1) {
    perror ("VIDIOCMCAPTURE(0)");
  }
  vmmap.frame = 1;
  
  for (;;){
    
    if (ioctl (ifid, VIDIOCMCAPTURE, &vmmap) == -1) {
      perror ("VIDIOCMCAPTURE()");
    }
    vmmap.frame = (vmmap.frame + 1) % vmbuf.frames;
    if (ioctl (ifid, VIDIOCSYNC, &vmmap.frame) == -1) {
      perror ("VIDIOCSYNC()");
    }
    
    /* process frame while the hardware captures next frame */
    
    if ((ofid = fopen(filename, "wb")) == NULL) {
      printf("Could not open file %s for writing\n", filename);
      return -1;
    }
    // The length / 8 of the JPEG is stored in the first 2 bytes of vmbuffer
    vmpointer = (unsigned short *)(vmbuffer+vmbuf.offsets[vmmap.frame]);
    count = (ssize_t)((unsigned int)(vmpointer[0])<<3);
    
    if (fwrite(vmbuffer+vmbuf.offsets[vmmap.frame] + 2, sizeof(unsigned char), (size_t)count, ofid) == count) {
      //	printf("Frame %d, %d bytes... success\n", frame, (int)count);
    }
    fclose(ofid);
    
    // jpeg read routine
    if ( (infile = fopen(filename, "rb") ) == NULL ){
      fprintf(stderr, "can't open %s\n", filename);
      return -1;
    }
    
    cinfo.err = jpeg_std_error(&(jerr.pub));
    jerr.pub.error_exit = my_error_exit;
    
    if (setjmp(jerr.setjmp_buffer)) {
      jpeg_destroy_decompress(&(cinfo));
      return -1;
    }
    
    jpeg_create_decompress(&cinfo);
    jpeg_stdio_src(&cinfo, infile);
    jpeg_read_header(&cinfo, TRUE);
    jpeg_start_decompress(&cinfo);
    
    
    row_stride = cinfo.output_width * cinfo.output_components;
    buffer = cinfo.mem->alloc_sarray( (j_common_ptr) &cinfo, JPOOL_IMAGE, row_stride, 1 );
    
    width  = cinfo.image_width;
    height = cinfo.image_height;
    
    current_line = cinfo.output_scanline;  
    last_line = cinfo.output_height;
    
    while (current_line < last_line) {
      
      jpeg_read_scanlines(&cinfo, buffer, 1);
      current_line = cinfo.output_scanline;     
      
      src = (unsigned char*)(buffer[0]);
      
      switch ( cinfo.out_color_space ){
      case JCS_RGB:                 // RGB 24bit
	for( i = 0; i < width; i++  ) {
	  r = *src++;
	  g = *src++;
	  b = *src++;
	  dst = fb + (height - current_line) + i * 480 ;
	  *dst = ((( r >> 3) << 11)|
		  (( g >> 2) << 5) |
		  (( b >> 3) << 0))  ;
	}
	break;
      case JCS_GRAYSCALE:                 //gray scale 8bit
	for( i = 0; i < width; i++ ) {
	  r = *src;
	  g = *src;
	  b = *src++;
	  dst = fb + (height - current_line) + i *480 ;
	  *dst = ((( r >> 3) << 11)|
		  (( g >> 2) << 5) |
		  (( b >> 3) << 0))  ;
	  
	}
	break;
      default:
	break;
      }
      
    }
    src = 0;
    
    (void) jpeg_finish_decompress(&cinfo); 
    jpeg_destroy_decompress(&cinfo);
    
    fclose(infile);
    infile = 0;
    
  }
  
  munmap (vmbuffer, vmbuf.size);
  
  close(ifid);
  
  close(console_fd);

  return 0;
}
